#
# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
#

from abc import ABC, abstractmethod
from pathlib import Path


class FileWriter(ABC):
    """
    Base File writer class
    """

    @abstractmethod
    def write(self, file_path: Path, content: bytes) -> int:
        """
        Writes the file to the specified location
        """
        ...



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from dataclasses import InitVar, dataclass
from typing import Any, Iterable, List, Mapping, MutableMapping, Optional, Union

from airbyte_cdk.sources.declarative.interpolation.interpolated_string import InterpolatedString
from airbyte_cdk.sources.declarative.partition_routers.partition_router import PartitionRouter
from airbyte_cdk.sources.declarative.requesters.request_option import (
    RequestOption,
    RequestOptionType,
)
from airbyte_cdk.sources.types import Config, StreamSlice, StreamState


@dataclass
class ListPartitionRouter(PartitionRouter):
    """
    Partition router that iterates over the values of a list
    If values is a string, then evaluate it as literal and assert the resulting literal is a list

    Attributes:
        values (Union[str, List[str]]): The values to iterate over
        cursor_field (Union[InterpolatedString, str]): The name of the cursor field
        config (Config): The user-provided configuration as specified by the source's spec
        request_option (Optional[RequestOption]): The request option to configure the HTTP request
    """

    values: Union[str, List[str]]
    cursor_field: Union[InterpolatedString, str]
    config: Config
    parameters: InitVar[Mapping[str, Any]]
    request_option: Optional[RequestOption] = None

    def __post_init__(self, parameters: Mapping[str, Any]) -> None:
        if isinstance(self.values, str):
            self.values = InterpolatedString.create(self.values, parameters=parameters).eval(
                self.config
            )
        self._cursor_field = (
            InterpolatedString(string=self.cursor_field, parameters=parameters)
            if isinstance(self.cursor_field, str)
            else self.cursor_field
        )

        self._cursor = None

    def get_request_params(
        self,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> Mapping[str, Any]:
        # Pass the stream_slice from the argument, not the cursor because the cursor is updated after processing the response
        return self._get_request_option(RequestOptionType.request_parameter, stream_slice)

    def get_request_headers(
        self,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> Mapping[str, Any]:
        # Pass the stream_slice from the argument, not the cursor because the cursor is updated after processing the response
        return self._get_request_option(RequestOptionType.header, stream_slice)

    def get_request_body_data(
        self,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> Mapping[str, Any]:
        # Pass the stream_slice from the argument, not the cursor because the cursor is updated after processing the response
        return self._get_request_option(RequestOptionType.body_data, stream_slice)

    def get_request_body_json(
        self,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> Mapping[str, Any]:
        # Pass the stream_slice from the argument, not the cursor because the cursor is updated after processing the response
        return self._get_request_option(RequestOptionType.body_json, stream_slice)

    def stream_slices(self) -> Iterable[StreamSlice]:
        return [
            StreamSlice(
                partition={self._cursor_field.eval(self.config): slice_value}, cursor_slice={}
            )
            for slice_value in self.values
        ]

    def _get_request_option(
        self, request_option_type: RequestOptionType, stream_slice: Optional[StreamSlice]
    ) -> Mapping[str, Any]:
        if (
            self.request_option
            and self.request_option.inject_into == request_option_type
            and stream_slice
        ):
            slice_value = stream_slice.get(self._cursor_field.eval(self.config))
            if slice_value:
                options: MutableMapping[str, Any] = {}
                self.request_option.inject_into_request(options, slice_value, self.config)
                return options
            else:
                return {}
        else:
            return {}

    def get_stream_state(self) -> Optional[Mapping[str, StreamState]]:
        """
        ListPartitionRouter doesn't have parent streams
        """
        pass



# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
"""CLI command for `airbyte-cdk`."""

USAGE = """CLI command for `airbyte-cdk`.

This CLI interface allows you to interact with your connector, including
testing and running commands.

**Basic Usage:**

```bash
airbyte-cdk --help
airbyte-cdk connector --help
airbyte-cdk manifest --help
```

**Running Statelessly:**

You can run the latest version of this CLI, from any machine, using `pipx` or `uvx`:

```bash
# Run the latest version of the CLI:
pipx run airbyte-cdk connector --help
uvx airbyte-cdk connector --help

# Run from a specific CDK version:
pipx run airbyte-cdk==6.5.1 connector --help
uvx airbyte-cdk==6.5.1 connector --help
```

**Running within your virtualenv:**

You can also run from your connector's virtualenv:

```bash
poetry run airbyte-cdk connector --help
```

"""

import os
import sys
from pathlib import Path
from types import ModuleType

import rich_click as click

from airbyte_cdk.test.standard_tests.util import create_connector_test_suite

# from airbyte_cdk.test.standard_tests import pytest_hooks
from airbyte_cdk.utils.connector_paths import (
    find_connector_root_from_name,
    resolve_connector_name_and_directory,
)

click.rich_click.TEXT_MARKUP = "markdown"

pytest: ModuleType | None
try:
    import pytest
except ImportError:
    pytest = None
    # Handle the case where pytest is not installed.
    # This prevents import errors when running the script without pytest installed.
    # We will raise an error later if pytest is required for a given command.


TEST_FILE_TEMPLATE = '''
# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
"""FAST Airbyte Standard Tests for the {connector_name} connector."""

#from airbyte_cdk.test.standard_tests import {base_class_name}
from airbyte_cdk.test.standard_tests.util import create_connector_test_suite
from pathlib import Path

pytest_plugins = [
    "airbyte_cdk.test.standard_tests.pytest_hooks",
]

TestSuite = create_connector_test_suite(
    connector_directory=Path(),
)

# Uncomment the following lines to create a custom test suite class:
#
# class TestSuite({base_class_name}):
#     """Test suite for the `{connector_name}` connector.
#
#     This class inherits from `{base_class_name}` and implements all of the tests in the suite.
#
#     As long as the class name starts with "Test", pytest will automatically discover and run the
#     tests in this class.
#     """
'''


@click.group(
    name="connector",
    help=__doc__.replace("\n", "\n\n"),  # Render docstring as help text (markdown)
)
def connector_cli_group() -> None:
    """Connector related commands."""
    pass


@connector_cli_group.command("test")
@click.argument(
    "connector",
    required=False,
    type=str,
    metavar="[CONNECTOR]",
)
@click.option(
    "--collect-only",
    is_flag=True,
    default=False,
    help="Only collect tests, do not run them.",
)
@click.option(
    "--pytest-arg",
    "pytest_args",  # ← map --pytest-arg into pytest_args
    type=str,
    multiple=True,
    help="Additional argument(s) to pass to pytest. Can be specified multiple times.",
)
@click.option(
    "--no-creds",
    is_flag=True,
    default=False,
    help="Skip tests that require credentials (marked with 'requires_creds').",
)
def connector_test(
    connector: str | Path | None = None,
    *,
    collect_only: bool = False,
    pytest_args: list[str] | None = None,
    no_creds: bool = False,
) -> None:
    """Run connector tests.

    This command runs the standard connector tests for a specific connector.

    [CONNECTOR] can be a connector name (e.g. 'source-pokeapi'), a path to a connector directory, or omitted to use the current working directory.
    If a string containing '/' is provided, it is treated as a path. Otherwise, it is treated as a connector name.

    If no connector name or directory is provided, we will look within the current working
    directory. If the current working directory is not a connector directory (e.g. starting
    with 'source-') and no connector name or path is provided, the process will fail.
    """
    click.echo("Connector test command executed.")
    connector_name, connector_directory = resolve_connector_name_and_directory(connector)

    pytest_args = pytest_args or []
    if collect_only:
        pytest_args.append("--collect-only")

    if no_creds:
        pytest_args.extend(["-m", "not requires_creds"])

    run_connector_tests(
        connector_name=connector_name,
        connector_directory=connector_directory,
        extra_pytest_args=pytest_args,
    )


def run_connector_tests(
    connector_name: str,
    connector_directory: Path,
    extra_pytest_args: list[str],
) -> None:
    if pytest is None:
        raise ImportError(
            "pytest is not installed. Please install pytest to run the connector tests."
        )

    connector_test_suite = create_connector_test_suite(
        connector_name=connector_name if not connector_directory else None,
        connector_directory=connector_directory,
    )

    pytest_args: list[str] = ["-p", "airbyte_cdk.test.standard_tests.pytest_hooks"]
    if connector_directory:
        pytest_args.append(f"--rootdir={connector_directory}")
        os.chdir(str(connector_directory))
    else:
        print("No connector directory provided. Running tests in the current directory.")

    file_text = TEST_FILE_TEMPLATE.format(
        base_class_name=connector_test_suite.__bases__[0].__name__,
        connector_name=connector_name,
    )
    test_file_path = Path() / ".tmp" / "integration_tests/test_airbyte_standards.py"
    test_file_path = test_file_path.resolve().absolute()
    test_file_path.parent.mkdir(parents=True, exist_ok=True)
    test_file_path.write_text(file_text)

    if extra_pytest_args:
        pytest_args.extend(extra_pytest_args)

    pytest_args.append(str(test_file_path))

    test_results_dir = connector_directory / "build" / "test-results"
    test_results_dir.mkdir(parents=True, exist_ok=True)
    junit_xml_path = test_results_dir / "standard-tests-junit.xml"
    pytest_args.extend(["--junitxml", str(junit_xml_path)])

    click.echo(f"Running tests from connector directory: {connector_directory}...")
    click.echo(f"Test file: {test_file_path}")
    click.echo(f"Pytest args: {pytest_args}")
    click.echo("Invoking Pytest...")
    exit_code = pytest.main(
        pytest_args,
        plugins=[],
    )
    sys.exit(exit_code)


__all__ = [
    "connector_cli_group",
]



import os
from typing import Optional

import jwt
from fastapi import HTTPException, Security, status
from fastapi.security import HTTPAuthorizationCredentials, HTTPBearer

security = HTTPBearer(auto_error=False)


def verify_jwt_token(
    credentials: Optional[HTTPAuthorizationCredentials] = Security(security),
) -> None:
    """
    Verify JWT token if AB_JWT_SIGNATURE_SECRET is set, otherwise allow through.

    Args:
        credentials: Bearer token credentials from request header

    Raises:
        HTTPException: If token is invalid or missing when secret is configured
    """
    jwt_secret = os.getenv("AB_JWT_SIGNATURE_SECRET")

    # If no secret is configured, allow all requests through
    if not jwt_secret:
        return

    if not credentials:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Bearer token required",
            headers={"WWW-Authenticate": "Bearer"},
        )

    try:
        jwt.decode(credentials.credentials, jwt_secret, algorithms=["HS256"])
    except jwt.InvalidTokenError:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Invalid token",
            headers={"WWW-Authenticate": "Bearer"},
        )



#
# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
#

import json
import logging
from collections import defaultdict
from dataclasses import InitVar, dataclass, field
from functools import partial
from typing import (
    Any,
    Callable,
    Iterable,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Set,
    Tuple,
    Union,
)

import requests
from typing_extensions import deprecated

from airbyte_cdk.sources.declarative.extractors.http_selector import HttpSelector
from airbyte_cdk.sources.declarative.interpolation import InterpolatedString
from airbyte_cdk.sources.declarative.partition_routers.single_partition_router import (
    SinglePartitionRouter,
)
from airbyte_cdk.sources.declarative.requesters.paginators.no_pagination import NoPagination
from airbyte_cdk.sources.declarative.requesters.paginators.paginator import Paginator
from airbyte_cdk.sources.declarative.requesters.query_properties import QueryProperties
from airbyte_cdk.sources.declarative.requesters.request_options import (
    DefaultRequestOptionsProvider,
    RequestOptionsProvider,
)
from airbyte_cdk.sources.declarative.requesters.requester import Requester
from airbyte_cdk.sources.declarative.retrievers.pagination_tracker import PaginationTracker
from airbyte_cdk.sources.declarative.retrievers.retriever import Retriever
from airbyte_cdk.sources.declarative.stream_slicers.stream_slicer import StreamSlicer
from airbyte_cdk.sources.source import ExperimentalClassWarning
from airbyte_cdk.sources.streams.core import StreamData
from airbyte_cdk.sources.streams.http.pagination_reset_exception import (
    PaginationResetRequiredException,
)
from airbyte_cdk.sources.types import Config, Record, StreamSlice
from airbyte_cdk.utils.mapping_helpers import combine_mappings

FULL_REFRESH_SYNC_COMPLETE_KEY = "__ab_full_refresh_sync_complete"
LOGGER = logging.getLogger("airbyte")


@dataclass
class SimpleRetriever(Retriever):
    """
    Retrieves records by synchronously sending requests to fetch records.

    The retriever acts as an orchestrator between the requester, the record selector, the paginator, and the stream slicer.

    For each stream slice, submit requests until there are no more pages of records to fetch.

    This retriever currently inherits from HttpStream to reuse the request submission and pagination machinery.
    As a result, some of the parameters passed to some methods are unused.
    The two will be decoupled in a future release.

    Attributes:
        stream_name (str): The stream's name
        stream_primary_key (Optional[Union[str, List[str], List[List[str]]]]): The stream's primary key
        requester (Requester): The HTTP requester
        record_selector (HttpSelector): The record selector
        paginator (Optional[Paginator]): The paginator
        stream_slicer (Optional[StreamSlicer]): The stream slicer
        parameters (Mapping[str, Any]): Additional runtime parameters to be used for string interpolation
    """

    requester: Requester
    record_selector: HttpSelector
    config: Config
    parameters: InitVar[Mapping[str, Any]]
    name: str
    _name: Union[InterpolatedString, str] = field(init=False, repr=False, default="")
    primary_key: Optional[Union[str, List[str], List[List[str]]]]
    _primary_key: str = field(init=False, repr=False, default="")
    paginator: Optional[Paginator] = None
    stream_slicer: StreamSlicer = field(
        default_factory=lambda: SinglePartitionRouter(parameters={})
    )
    request_option_provider: RequestOptionsProvider = field(
        default_factory=lambda: DefaultRequestOptionsProvider(parameters={})
    )
    ignore_stream_slicer_parameters_on_paginated_requests: bool = False
    additional_query_properties: Optional[QueryProperties] = None
    log_formatter: Optional[Callable[[requests.Response], Any]] = None
    pagination_tracker_factory: Callable[[], PaginationTracker] = field(
        default_factory=lambda: lambda: PaginationTracker()
    )

    def __post_init__(self, parameters: Mapping[str, Any]) -> None:
        self._paginator = self.paginator or NoPagination(parameters=parameters)
        self._parameters = parameters
        self._name = (
            InterpolatedString(self._name, parameters=parameters)
            if isinstance(self._name, str)
            else self._name
        )

    @property  # type: ignore
    def name(self) -> str:
        """
        :return: Stream name
        """
        return (
            str(self._name.eval(self.config))
            if isinstance(self._name, InterpolatedString)
            else self._name
        )

    @name.setter
    def name(self, value: str) -> None:
        if not isinstance(value, property):
            self._name = value

    def _get_mapping(
        self, method: Callable[..., Optional[Union[Mapping[str, Any], str]]], **kwargs: Any
    ) -> Tuple[Union[Mapping[str, Any], str], Set[str]]:
        """
        Get mapping from the provided method, and get the keys of the mapping.
        If the method returns a string, it will return the string and an empty set.
        If the method returns a dict, it will return the dict and its keys.
        """
        mapping = method(**kwargs) or {}
        keys = set(mapping.keys()) if not isinstance(mapping, str) else set()
        return mapping, keys

    def _get_request_options(
        self,
        stream_slice: Optional[StreamSlice],
        next_page_token: Optional[Mapping[str, Any]],
        paginator_method: Callable[..., Optional[Union[Mapping[str, Any], str]]],
        stream_slicer_method: Callable[..., Optional[Union[Mapping[str, Any], str]]],
    ) -> Union[Mapping[str, Any], str]:
        """
        Get the request_option from the paginator and the stream slicer.
        Raise a ValueError if there's a key collision
        Returned merged mapping otherwise
        """
        is_body_json = paginator_method.__name__ == "get_request_body_json"

        mappings = [
            paginator_method(
                stream_slice=stream_slice,
                next_page_token=next_page_token,
            ),
        ]
        if not next_page_token or not self.ignore_stream_slicer_parameters_on_paginated_requests:
            mappings.append(
                stream_slicer_method(
                    stream_slice=stream_slice,
                    next_page_token=next_page_token,
                )
            )
        return combine_mappings(mappings, allow_same_value_merge=is_body_json)

    def _request_headers(
        self,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> Mapping[str, Any]:
        """
        Specifies request headers.
        Authentication headers will overwrite any overlapping headers returned from this method.
        """
        headers = self._get_request_options(
            stream_slice,
            next_page_token,
            self._paginator.get_request_headers,
            self.request_option_provider.get_request_headers,
        )
        if isinstance(headers, str):
            raise ValueError("Request headers cannot be a string")
        return {str(k): str(v) for k, v in headers.items()}

    def _request_params(
        self,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> Mapping[str, Any]:
        """
        Specifies the query parameters that should be set on an outgoing HTTP request given the inputs.

        E.g: you might want to define query parameters for paging if next_page_token is not None.
        """
        params = self._get_request_options(
            stream_slice,
            next_page_token,
            self._paginator.get_request_params,
            self.request_option_provider.get_request_params,
        )
        if isinstance(params, str):
            raise ValueError("Request params cannot be a string")
        return params

    def _request_body_data(
        self,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> Union[Mapping[str, Any], str]:
        """
        Specifies how to populate the body of the request with a non-JSON payload.

        If returns a ready text that it will be sent as is.
        If returns a dict that it will be converted to a urlencoded form.
        E.g. {"key1": "value1", "key2": "value2"} => "key1=value1&key2=value2"

        At the same time only one of the 'request_body_data' and 'request_body_json' functions can be overridden.
        """
        return self._get_request_options(
            stream_slice,
            next_page_token,
            self._paginator.get_request_body_data,
            self.request_option_provider.get_request_body_data,
        )

    def _request_body_json(
        self,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> Optional[Mapping[str, Any]]:
        """
        Specifies how to populate the body of the request with a JSON payload.

        At the same time only one of the 'request_body_data' and 'request_body_json' functions can be overridden.
        """
        body_json = self._get_request_options(
            stream_slice,
            next_page_token,
            self._paginator.get_request_body_json,
            self.request_option_provider.get_request_body_json,
        )
        if isinstance(body_json, str):
            raise ValueError("Request body json cannot be a string")
        return body_json

    def _paginator_path(
        self,
        next_page_token: Optional[Mapping[str, Any]] = None,
        stream_slice: Optional[StreamSlice] = None,
    ) -> Optional[str]:
        """
        If the paginator points to a path, follow it, else return nothing so the requester is used.
        :param next_page_token:
        :return:
        """
        return self._paginator.path(
            next_page_token=next_page_token,
            stream_state={},  # stream_state as an interpolation context is deprecated
            stream_slice=stream_slice,
        )

    def _parse_response(
        self,
        response: Optional[requests.Response],
        records_schema: Mapping[str, Any],
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> Iterable[Record]:
        if not response:
            yield from []
        else:
            yield from self.record_selector.select_records(
                response=response,
                stream_state={},  # stream_state as an interpolation context is deprecated
                records_schema=records_schema,
                stream_slice=stream_slice,
                next_page_token=next_page_token,
            )

    @property  # type: ignore
    def primary_key(self) -> Optional[Union[str, List[str], List[List[str]]]]:
        """The stream's primary key"""
        return self._primary_key

    @primary_key.setter
    def primary_key(self, value: str) -> None:
        if not isinstance(value, property):
            self._primary_key = value

    def _next_page_token(
        self,
        response: requests.Response,
        last_page_size: int,
        last_record: Optional[Record],
        last_page_token_value: Optional[Any],
    ) -> Optional[Mapping[str, Any]]:
        """
        Specifies a pagination strategy.

        The value returned from this method is passed to most other methods in this class. Use it to form a request e.g: set headers or query params.

        :return: The token for the next page from the input response object. Returning None means there are no more pages to read in this response.
        """
        return self._paginator.next_page_token(
            response=response,
            last_page_size=last_page_size,
            last_record=last_record,
            last_page_token_value=last_page_token_value,
        )

    def _fetch_next_page(
        self,
        stream_slice: StreamSlice,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> Optional[requests.Response]:
        return self.requester.send_request(
            path=self._paginator_path(
                next_page_token=next_page_token,
                stream_slice=stream_slice,
            ),
            stream_state={},  # stream_state as an interpolation context is deprecated
            stream_slice=stream_slice,
            next_page_token=next_page_token,
            request_headers=self._request_headers(
                stream_slice=stream_slice,
                next_page_token=next_page_token,
            ),
            request_params=self._request_params(
                stream_slice=stream_slice,
                next_page_token=next_page_token,
            ),
            request_body_data=self._request_body_data(
                stream_slice=stream_slice,
                next_page_token=next_page_token,
            ),
            request_body_json=self._request_body_json(
                stream_slice=stream_slice,
                next_page_token=next_page_token,
            ),
            log_formatter=self.log_formatter,
        )

    # This logic is similar to _read_pages in the HttpStream class. When making changes here, consider making changes there as well.
    def _read_pages(
        self,
        records_generator_fn: Callable[[Optional[requests.Response]], Iterable[Record]],
        stream_slice: StreamSlice,
    ) -> Iterable[Record]:
        original_stream_slice = stream_slice
        pagination_tracker = self.pagination_tracker_factory()
        reset_pagination = False
        next_page_token = self._get_initial_next_page_token()
        while True:
            merged_records: MutableMapping[str, Any] = defaultdict(dict)
            last_page_size = 0
            last_record: Optional[Record] = None

            response = None
            try:
                if self.additional_query_properties:
                    for (
                        properties
                    ) in self.additional_query_properties.get_request_property_chunks():
                        stream_slice = StreamSlice(
                            partition=stream_slice.partition or {},
                            cursor_slice=stream_slice.cursor_slice or {},
                            extra_fields={"query_properties": properties},
                        )
                        response = self._fetch_next_page(stream_slice, next_page_token)

                        for current_record in records_generator_fn(response):
                            if self.additional_query_properties.property_chunking:
                                merge_key = self.additional_query_properties.property_chunking.get_merge_key(
                                    current_record
                                )
                                if merge_key:
                                    _deep_merge(merged_records[merge_key], current_record)
                                else:
                                    # We should still emit records even if the record did not have a merge key
                                    pagination_tracker.observe(current_record)
                                    last_page_size += 1
                                    last_record = current_record
                                    yield current_record
                            else:
                                pagination_tracker.observe(current_record)
                                last_page_size += 1
                                last_record = current_record
                                yield current_record

                    for merged_record in merged_records.values():
                        record = Record(
                            data=merged_record, stream_name=self.name, associated_slice=stream_slice
                        )
                        pagination_tracker.observe(record)
                        last_page_size += 1
                        last_record = record
                        yield record
                else:
                    response = self._fetch_next_page(stream_slice, next_page_token)
                    for current_record in records_generator_fn(response):
                        pagination_tracker.observe(current_record)
                        last_page_size += 1
                        last_record = current_record
                        yield current_record
            except PaginationResetRequiredException:
                reset_pagination = True
            else:
                if not response:
                    break

            if reset_pagination or pagination_tracker.has_reached_limit():
                next_page_token = self._get_initial_next_page_token()
                previous_slice = stream_slice
                stream_slice = pagination_tracker.reduce_slice_range_if_possible(
                    stream_slice, original_stream_slice
                )
                LOGGER.info(
                    f"Hitting PaginationReset event. StreamSlice used will go from {previous_slice} to {stream_slice}"
                )
                reset_pagination = False
            else:
                last_page_token_value = (
                    next_page_token.get("next_page_token") if next_page_token else None
                )
                next_page_token = self._next_page_token(
                    response=response,  # type:ignore # we are breaking from the loop on the try/else if there are no response so this should be fine
                    last_page_size=last_page_size,
                    last_record=last_record,
                    last_page_token_value=last_page_token_value,
                )
                if not next_page_token:
                    break

        # Always return an empty generator just in case no records were ever yielded
        yield from []

    def _get_initial_next_page_token(self) -> Optional[Mapping[str, Any]]:
        initial_token = self._paginator.get_initial_token()
        next_page_token = {"next_page_token": initial_token} if initial_token is not None else None
        return next_page_token

    def read_records(
        self,
        records_schema: Mapping[str, Any],
        stream_slice: Optional[StreamSlice] = None,
    ) -> Iterable[StreamData]:
        """
        Fetch a stream's records from an HTTP API source

        :param records_schema: json schema to describe record
        :param stream_slice: The stream slice to read data for
        :return: The records read from the API source
        """
        _slice = stream_slice or StreamSlice(partition={}, cursor_slice={})  # None-check

        record_generator = partial(
            self._parse_records,
            stream_slice=stream_slice,
            records_schema=records_schema,
        )
        yield from self._read_pages(record_generator, _slice)

    def _parse_records(
        self,
        response: Optional[requests.Response],
        records_schema: Mapping[str, Any],
        stream_slice: Optional[StreamSlice],
    ) -> Iterable[Record]:
        yield from self._parse_response(
            response,
            stream_slice=stream_slice,
            records_schema=records_schema,
        )

    def must_deduplicate_query_params(self) -> bool:
        return True

    @staticmethod
    def _to_partition_key(to_serialize: Any) -> str:
        # separators have changed in Python 3.4. To avoid being impacted by further change, we explicitly specify our own value
        return json.dumps(to_serialize, indent=None, separators=(",", ":"), sort_keys=True)


def _deep_merge(
    target: MutableMapping[str, Any], source: Union[Record, MutableMapping[str, Any]]
) -> None:
    """
    Recursively merge two dictionaries, combining nested dictionaries instead of overwriting them.

    :param target: The dictionary to merge into (modified in place)
    :param source: The dictionary to merge from
    """
    for key, value in source.items():
        if (
            key in target
            and isinstance(target[key], MutableMapping)
            and isinstance(value, MutableMapping)
        ):
            _deep_merge(target[key], value)
        else:
            target[key] = value


@deprecated(
    "This class is experimental. Use at your own risk.",
    category=ExperimentalClassWarning,
)
@dataclass
class LazySimpleRetriever(SimpleRetriever):
    """
    A retriever that supports lazy loading from parent streams.
    """

    def _read_pages(
        self,
        records_generator_fn: Callable[[Optional[requests.Response]], Iterable[Record]],
        stream_slice: StreamSlice,
    ) -> Iterable[Record]:
        response = stream_slice.extra_fields["child_response"]
        if response:
            last_page_size, last_record = 0, None
            for record in records_generator_fn(response):  # type: ignore[call-arg] # only _parse_records expected as a func
                last_page_size += 1
                last_record = record
                yield record

            next_page_token = self._next_page_token(response, last_page_size, last_record, None)
            if next_page_token:
                yield from self._paginate(
                    next_page_token,
                    records_generator_fn,
                    stream_slice,
                )

            yield from []
        else:
            # coderabbit detected an interesting bug/gap where if we were to not get a child_response, we
            # might recurse forever. This might not be the case, but it is worth noting that this code path
            # isn't comprehensively tested.
            yield from self._read_pages(records_generator_fn, stream_slice)

    def _paginate(
        self,
        next_page_token: Any,
        records_generator_fn: Callable[[Optional[requests.Response]], Iterable[Record]],
        stream_slice: StreamSlice,
    ) -> Iterable[Record]:
        """Handle pagination by fetching subsequent pages."""
        pagination_complete = False

        while not pagination_complete:
            response = self._fetch_next_page(stream_slice, next_page_token)
            last_page_size, last_record = 0, None

            for record in records_generator_fn(response):  # type: ignore[call-arg] # only _parse_records expected as a func
                last_page_size += 1
                last_record = record
                yield record

            if not response:
                pagination_complete = True
            else:
                last_page_token_value = (
                    next_page_token.get("next_page_token") if next_page_token else None
                )
                next_page_token = self._next_page_token(
                    response, last_page_size, last_record, last_page_token_value
                )

                if not next_page_token:
                    pagination_complete = True



# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
"""Main entry point for the Airbyte Manifest Server server."""

import uvicorn


def run_server(
    host: str = "127.0.0.1", port: int = 8000, reload: bool = False, log_level: str = "info"
) -> None:
    """Run the FastAPI server."""

    print(f"🚀 Starting Airbyte CDK Manifest Server on {host}:{port}")

    uvicorn.run(
        "airbyte_cdk.manifest_server.app:app",
        host=host,
        port=port,
        reload=reload,
        log_level=log_level,
    )


if __name__ == "__main__":
    run_server()



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from abc import ABC, abstractmethod
from typing import Any, Optional

from airbyte_cdk.sources.types import Record, StreamSlice, StreamState


class Cursor(ABC):
    """
    Cursors are components that allow for checkpointing the current state of a sync. They keep track of what data has been consumed
    and allows for syncs to be resumed from a specific point based on that information.
    """

    @abstractmethod
    def set_initial_state(self, stream_state: StreamState) -> None:
        """
        Cursors are not initialized with their state. As state is needed in order to function properly, this method should be called
        before calling anything else

        :param stream_state: The state of the stream as returned by get_stream_state
        """

    def observe(self, stream_slice: StreamSlice, record: Record) -> None:
        """
        Register a record with the cursor; the cursor instance can then use it to manage the state of the in-progress stream read.

        :param stream_slice: The current slice, which may or may not contain the most recently observed record
        :param record: the most recently-read record, which the cursor can use to update the stream state. Outwardly-visible changes to the
          stream state may need to be deferred depending on whether the source reliably orders records by the cursor field.
        """
        pass

    @abstractmethod
    def close_slice(self, stream_slice: StreamSlice, *args: Any) -> None:
        """
        Update state based on the stream slice. Note that `stream_slice.cursor_slice` and `most_recent_record.associated_slice` are expected
        to be the same but we make it explicit here that `stream_slice` should be leveraged to update the state. We do not pass in the
        latest record, since cursor instances should maintain the relevant internal state on their own.

        :param stream_slice: slice to close
        """

    @abstractmethod
    def get_stream_state(self) -> StreamState:
        """
        Returns the current stream state. We would like to restrict it's usage since it does expose internal of state. As of 2023-06-14, it
        is used for two things:
        * Interpolation of the requests
        * Transformation of records
        * Saving the state

        For the first case, we are probably stuck with exposing the stream state. For the second, we can probably expose a method that
        allows for emitting the state to the platform.
        """

    @abstractmethod
    def should_be_synced(self, record: Record) -> bool:
        """
        Evaluating if a record should be synced allows for filtering and stop condition on pagination
        """

    @abstractmethod
    def select_state(self, stream_slice: Optional[StreamSlice] = None) -> Optional[StreamState]:
        """
        Get the state value of a specific stream_slice. For incremental or resumable full refresh cursors which only manage state in
        a single dimension this is the entire state object. For per-partition cursors used by substreams, this returns the state of
        a specific parent delineated by the incoming slice's partition object.
        """



# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
"""Catalog provider implementation.

A catalog provider wraps a configured catalog and configured streams. This class is responsible for
providing information about the catalog and streams. A catalog provider can also be updated with new
streams as they are discovered, providing a thin layer of abstraction over the configured catalog.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, cast, final

from airbyte_cdk.models import ConfiguredAirbyteCatalog
from airbyte_cdk.sql import exceptions as exc
from airbyte_cdk.sql._util.name_normalizers import LowerCaseNormalizer

if TYPE_CHECKING:
    from airbyte_cdk.models import ConfiguredAirbyteStream


class CatalogProvider:
    """A catalog provider wraps a configured catalog and configured streams.

    This class is responsible for providing information about the catalog and streams.

    Note:
    - The catalog provider is not responsible for managing the catalog or streams but it may
      be updated with new streams as they are discovered.
    """

    def __init__(
        self,
        configured_catalog: ConfiguredAirbyteCatalog,
    ) -> None:
        """Initialize the catalog manager with a catalog object reference.

        Since the catalog is passed by reference, the catalog manager may be updated with new
        streams as they are discovered.
        """
        self._catalog: ConfiguredAirbyteCatalog = self.validate_catalog(configured_catalog)

    @staticmethod
    def validate_catalog(catalog: ConfiguredAirbyteCatalog) -> Any:
        """Validate the catalog to ensure it is valid.

        This requires ensuring that `generationId` and `minGenerationId` are both set. If
        not, both values will be set to `1`.
        """
        for stream in catalog.streams:
            if stream.generation_id is None:
                stream.generation_id = 1
            if stream.minimum_generation_id is None:
                stream.minimum_generation_id = 1
            if stream.sync_id is None:
                stream.sync_id = 1  # This should ideally increment monotonically with each sync.

        return catalog

    @property
    def configured_catalog(self) -> ConfiguredAirbyteCatalog:
        """Return the configured catalog."""
        return self._catalog

    @property
    def stream_names(self) -> list[str]:
        """Return the names of the streams in the catalog."""
        return list({stream.stream.name for stream in self.configured_catalog.streams})

    def get_configured_stream_info(
        self,
        stream_name: str,
    ) -> ConfiguredAirbyteStream:
        """Return the column definitions for the given stream."""
        if not self.configured_catalog:
            raise exc.AirbyteInternalError(
                message="Cannot get stream JSON schema without a catalog.",
            )

        matching_streams: list[ConfiguredAirbyteStream] = [
            stream
            for stream in self.configured_catalog.streams
            if stream.stream.name == stream_name
        ]
        if not matching_streams:
            raise exc.AirbyteStreamNotFoundError(
                stream_name=stream_name,
                context={
                    "available_streams": [
                        stream.stream.name for stream in self.configured_catalog.streams
                    ],
                },
            )

        if len(matching_streams) > 1:
            raise exc.AirbyteInternalError(
                message="Multiple streams found with same name.",
                context={
                    "stream_name": stream_name,
                },
            )

        return matching_streams[0]

    @final
    def get_stream_json_schema(
        self,
        stream_name: str,
    ) -> dict[str, Any]:
        """Return the column definitions for the given stream."""
        return cast(dict[str, Any], self.get_configured_stream_info(stream_name).stream.json_schema)

    def get_stream_properties(
        self,
        stream_name: str,
    ) -> dict[str, dict[str, Any]]:
        """Return the names of the top-level properties for the given stream."""
        return cast(dict[str, Any], self.get_stream_json_schema(stream_name)["properties"])

    def get_primary_keys(
        self,
        stream_name: str,
    ) -> list[str] | None:
        """Return the primary key column names for the given stream.

        We return `source_defined_primary_key` if set, or `primary_key` otherwise. If both are set,
        we assume they should not should differ, since Airbyte data integrity constraints do not
        permit overruling a source's pre-defined primary keys. If neither is set, we return `None`.

        Returns:
            A list of column names that constitute the primary key, or None if no primary key is defined.
        """
        configured_stream = self.get_configured_stream_info(stream_name)
        pks = configured_stream.stream.source_defined_primary_key or configured_stream.primary_key

        if not pks:
            return None

        normalized_pks: list[list[str]] = [
            [LowerCaseNormalizer.normalize(c) for c in pk] for pk in pks
        ]

        for pk_nodes in normalized_pks:
            if len(pk_nodes) != 1:
                raise exc.AirbyteError(
                    message=(
                        "Nested primary keys are not supported. "
                        "Each PK column should have exactly one node. "
                    ),
                    context={
                        "stream_name": stream_name,
                        "primary_key_nodes": pk_nodes,
                    },
                )

        return [pk_nodes[0] for pk_nodes in normalized_pks]



"""
Stream-related API models.

These models define the structure for stream reading operations and responses.
They accurately reflect the runtime types returned by the CDK, particularly
fixing type mismatches like slice_descriptor being a string rather than an object.
"""

from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel


class HttpRequest(BaseModel):
    """HTTP request details."""

    url: str
    headers: Optional[Dict[str, Any]]
    http_method: str
    body: Optional[str] = None


class HttpResponse(BaseModel):
    """HTTP response details."""

    status: int
    body: Optional[str] = None
    headers: Optional[Dict[str, Any]] = None


class LogMessage(BaseModel):
    """Log message from stream processing."""

    message: str
    level: str
    internal_message: Optional[str] = None
    stacktrace: Optional[str] = None


class AuxiliaryRequest(BaseModel):
    """Auxiliary HTTP request made during stream processing."""

    title: str
    type: str
    description: str
    request: HttpRequest
    response: HttpResponse


class StreamReadPages(BaseModel):
    """Pages of data read from a stream slice."""

    records: List[object]
    request: Optional[HttpRequest] = None
    response: Optional[HttpResponse] = None


class StreamReadSlices(BaseModel):
    """Slices of data read from a stream."""

    pages: List[StreamReadPages]
    slice_descriptor: Optional[Union[Dict[str, Any], str]]  # We're seeing strings at runtime
    state: Optional[List[Dict[str, Any]]] = None
    auxiliary_requests: Optional[List[AuxiliaryRequest]] = None


class StreamReadResponse(BaseModel):
    """Complete stream read response with properly typed fields."""

    logs: List[LogMessage]
    slices: List[StreamReadSlices]
    test_read_limit_reached: bool
    auxiliary_requests: List[AuxiliaryRequest]
    inferred_schema: Optional[Dict[str, Any]]
    inferred_datetime_formats: Optional[Dict[str, str]]
    latest_config_update: Optional[Dict[str, Any]]



"""Models to represent the structure of a `metadata.yaml` file."""

from __future__ import annotations

from enum import Enum
from pathlib import Path

import yaml
from pydantic import BaseModel, Field


class ConnectorLanguage(str, Enum):
    """Connector implementation language."""

    PYTHON = "python"
    JAVA = "java"
    LOW_CODE = "low-code"
    MANIFEST_ONLY = "manifest-only"
    UNKNOWN = "unknown"


class ConnectorBuildOptions(BaseModel):
    """Connector build options from metadata.yaml."""

    model_config = {"extra": "allow"}

    baseImage: str | None = Field(
        None,
        description="Base image to use for building the connector",
    )
    path: str | None = Field(
        None,
        description="Path to the connector code within the repository",
    )


class SuggestedStreams(BaseModel):
    """Suggested streams from metadata.yaml."""

    streams: list[str] = Field(
        default=[],
        description="List of suggested streams for the connector",
    )


class ConnectorMetadata(BaseModel):
    """Connector metadata from metadata.yaml."""

    model_config = {"extra": "allow"}

    dockerRepository: str = Field(..., description="Docker repository for the connector image")
    dockerImageTag: str = Field(..., description="Docker image tag for the connector")

    tags: list[str] = Field(
        default=[],
        description="List of tags for the connector",
    )

    suggestedStreams: SuggestedStreams | None = Field(
        default=None,
        description="Suggested streams for the connector",
    )

    @property
    def language(self) -> ConnectorLanguage:
        """Get the connector language."""
        for tag in self.tags:
            if tag.startswith("language:"):
                language = tag.split(":", 1)[1]
                if language == "python":
                    return ConnectorLanguage.PYTHON
                elif language == "java":
                    return ConnectorLanguage.JAVA
                elif language == "low-code":
                    return ConnectorLanguage.LOW_CODE
                elif language == "manifest-only":
                    return ConnectorLanguage.MANIFEST_ONLY

        return ConnectorLanguage.UNKNOWN

    connectorBuildOptions: ConnectorBuildOptions | None = Field(
        None, description="Options for building the connector"
    )


class MetadataFile(BaseModel):
    """Represents the structure of a metadata.yaml file."""

    model_config = {"extra": "allow"}

    data: ConnectorMetadata = Field(..., description="Connector metadata")

    @classmethod
    def from_file(
        cls,
        file_path: Path,
    ) -> MetadataFile:
        """Load metadata from a YAML file."""
        if not file_path.exists():
            raise FileNotFoundError(f"Metadata file not found: {file_path!s}")

        metadata_content = file_path.read_text()
        metadata_dict = yaml.safe_load(metadata_content)

        if not metadata_dict or "data" not in metadata_dict:
            raise ValueError(
                "Invalid metadata format: missing 'data' field in YAML file '{file_path!s}'"
            )

        metadata_file = MetadataFile.model_validate(metadata_dict)
        return metadata_file



# Copyright (c) 2025 Airbyte, Inc., all rights reserved.

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional, Set


@dataclass
class PropertySelector(ABC):
    """
    Describes the interface for selecting and transforming properties from a configured stream's schema
    to determine which properties should be queried from the API.
    """

    @abstractmethod
    def select(self) -> Optional[Set[str]]:
        """
        Selects and returns the set of properties that should be queried from the API based on the
        configured stream's schema and any applicable transformations.

        Returns:
            Set[str]: The set of property names to query
        """
        pass



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import logging
from dataclasses import InitVar, dataclass, field
from typing import Any, Callable, Mapping, MutableMapping, Optional, Union
from urllib.parse import urljoin

import requests

from airbyte_cdk.sources.declarative.auth.declarative_authenticator import (
    DeclarativeAuthenticator,
    NoAuth,
)
from airbyte_cdk.sources.declarative.decoders import Decoder
from airbyte_cdk.sources.declarative.decoders.json_decoder import JsonDecoder
from airbyte_cdk.sources.declarative.interpolation.interpolated_string import (
    InterpolatedString,
)
from airbyte_cdk.sources.declarative.requesters.request_options.interpolated_request_options_provider import (
    InterpolatedRequestOptionsProvider,
)
from airbyte_cdk.sources.declarative.requesters.requester import HttpMethod, Requester
from airbyte_cdk.sources.message import MessageRepository, NoopMessageRepository
from airbyte_cdk.sources.streams.call_rate import APIBudget
from airbyte_cdk.sources.streams.http import HttpClient
from airbyte_cdk.sources.streams.http.error_handlers import ErrorHandler
from airbyte_cdk.sources.types import Config, EmptyString, StreamSlice, StreamState
from airbyte_cdk.utils.mapping_helpers import (
    combine_mappings,
    get_interpolation_context,
)


@dataclass
class HttpRequester(Requester):
    """
    Default implementation of a Requester

    Attributes:
        name (str): Name of the stream. Only used for request/response caching
        url_base (Union[InterpolatedString, str]): Base url to send requests to
        path (Union[InterpolatedString, str]): Path to send requests to
        http_method (Union[str, HttpMethod]): HTTP method to use when sending requests
        request_options_provider (Optional[InterpolatedRequestOptionsProvider]): request option provider defining the options to set on outgoing requests
        authenticator (DeclarativeAuthenticator): Authenticator defining how to authenticate to the source
        error_handler (Optional[ErrorHandler]): Error handler defining how to detect and handle errors
        backoff_strategies (Optional[List[BackoffStrategy]]): List of backoff strategies to use when retrying requests
        config (Config): The user-provided configuration as specified by the source's spec
        use_cache (bool): Indicates that data should be cached for this stream
    """

    name: str
    config: Config
    parameters: InitVar[Mapping[str, Any]]

    url: Optional[Union[InterpolatedString, str]] = None
    url_base: Optional[Union[InterpolatedString, str]] = None
    path: Optional[Union[InterpolatedString, str]] = None
    authenticator: Optional[DeclarativeAuthenticator] = None
    http_method: Union[str, HttpMethod] = HttpMethod.GET
    request_options_provider: Optional[InterpolatedRequestOptionsProvider] = None
    error_handler: Optional[ErrorHandler] = None
    api_budget: Optional[APIBudget] = None
    disable_retries: bool = False
    message_repository: MessageRepository = NoopMessageRepository()
    use_cache: bool = False
    _exit_on_rate_limit: bool = False
    stream_response: bool = False
    decoder: Decoder = field(default_factory=lambda: JsonDecoder(parameters={}))

    def __post_init__(self, parameters: Mapping[str, Any]) -> None:
        self._url = InterpolatedString.create(
            self.url if self.url else EmptyString, parameters=parameters
        )
        # deprecated
        self._url_base = InterpolatedString.create(
            self.url_base if self.url_base else EmptyString, parameters=parameters
        )
        # deprecated
        self._path = InterpolatedString.create(
            self.path if self.path else EmptyString, parameters=parameters
        )
        if self.request_options_provider is None:
            self._request_options_provider = InterpolatedRequestOptionsProvider(
                config=self.config, parameters=parameters
            )
        elif isinstance(self.request_options_provider, dict):
            self._request_options_provider = InterpolatedRequestOptionsProvider(
                config=self.config, **self.request_options_provider
            )
        else:
            self._request_options_provider = self.request_options_provider
        self._authenticator = self.authenticator or NoAuth(parameters=parameters)
        self._http_method = (
            HttpMethod[self.http_method] if isinstance(self.http_method, str) else self.http_method
        )
        self.error_handler = self.error_handler
        self._parameters = parameters

        if self.error_handler is not None and hasattr(self.error_handler, "backoff_strategies"):
            backoff_strategies = self.error_handler.backoff_strategies  # type: ignore
        else:
            backoff_strategies = None

        self._http_client = HttpClient(
            name=self.name,
            logger=self.logger,
            error_handler=self.error_handler,
            api_budget=self.api_budget,
            authenticator=self._authenticator,
            use_cache=self.use_cache,
            backoff_strategy=backoff_strategies,
            disable_retries=self.disable_retries,
            message_repository=self.message_repository,
        )

    @property
    def exit_on_rate_limit(self) -> bool:
        return self._exit_on_rate_limit

    @exit_on_rate_limit.setter
    def exit_on_rate_limit(self, value: bool) -> None:
        self._exit_on_rate_limit = value

    def get_authenticator(self) -> DeclarativeAuthenticator:
        return self._authenticator

    def get_url(
        self,
        *,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> str:
        interpolation_context = get_interpolation_context(
            stream_state=stream_state,
            stream_slice=stream_slice,
            next_page_token=next_page_token,
        )

        return str(self._url.eval(self.config, **interpolation_context))

    def _get_url(
        self,
        *,
        path: Optional[str] = None,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> str:
        url = self.get_url(
            stream_state=stream_state,
            stream_slice=stream_slice,
            next_page_token=next_page_token,
        )

        url_base = self.get_url_base(
            stream_state=stream_state,
            stream_slice=stream_slice,
            next_page_token=next_page_token,
        )

        path = path or self.get_path(
            stream_state=stream_state,
            stream_slice=stream_slice,
            next_page_token=next_page_token,
        )

        full_url = (
            self._join_url(url_base, path)
            if url_base
            else self._join_url(url, path)
            if path
            else url
        )

        return full_url

    def get_url_base(
        self,
        *,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> str:
        interpolation_context = get_interpolation_context(
            stream_state=stream_state,
            stream_slice=stream_slice,
            next_page_token=next_page_token,
        )
        return str(self._url_base.eval(self.config, **interpolation_context))

    def get_path(
        self,
        *,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> str:
        interpolation_context = get_interpolation_context(
            stream_state=stream_state,
            stream_slice=stream_slice,
            next_page_token=next_page_token,
        )
        path = str(self._path.eval(self.config, **interpolation_context))
        return path.lstrip("/")

    def get_method(self) -> HttpMethod:
        return self._http_method

    def get_request_params(
        self,
        *,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> MutableMapping[str, Any]:
        return self._request_options_provider.get_request_params(
            stream_state=stream_state,
            stream_slice=stream_slice,
            next_page_token=next_page_token,
        )

    def get_request_headers(
        self,
        *,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> Mapping[str, Any]:
        return self._request_options_provider.get_request_headers(
            stream_state=stream_state,
            stream_slice=stream_slice,
            next_page_token=next_page_token,
        )

    # fixing request options provider types has a lot of dependencies
    def get_request_body_data(  # type: ignore
        self,
        *,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> Union[Mapping[str, Any], str]:
        return (
            self._request_options_provider.get_request_body_data(
                stream_state=stream_state,
                stream_slice=stream_slice,
                next_page_token=next_page_token,
            )
            or {}
        )

    # fixing request options provider types has a lot of dependencies
    def get_request_body_json(  # type: ignore
        self,
        *,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> Optional[Mapping[str, Any]]:
        return self._request_options_provider.get_request_body_json(
            stream_state=stream_state,
            stream_slice=stream_slice,
            next_page_token=next_page_token,
        )

    @property
    def logger(self) -> logging.Logger:
        return logging.getLogger(f"airbyte.HttpRequester.{self.name}")

    def _get_request_options(
        self,
        stream_state: Optional[StreamState],
        stream_slice: Optional[StreamSlice],
        next_page_token: Optional[Mapping[str, Any]],
        requester_method: Callable[..., Optional[Union[Mapping[str, Any], str]]],
        auth_options_method: Callable[..., Optional[Union[Mapping[str, Any], str]]],
        extra_options: Optional[Union[Mapping[str, Any], str]] = None,
    ) -> Union[Mapping[str, Any], str]:
        """
        Get the request_option from the requester, the authenticator and extra_options passed in.
        Raise a ValueError if there's a key collision
        Returned merged mapping otherwise
        """

        is_body_json = requester_method.__name__ == "get_request_body_json"

        return combine_mappings(
            [
                requester_method(
                    stream_state=stream_state,
                    stream_slice=stream_slice,
                    next_page_token=next_page_token,
                ),
                auth_options_method(),
                extra_options,
            ],
            allow_same_value_merge=is_body_json,
        )

    def _request_headers(
        self,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
        extra_headers: Optional[Mapping[str, Any]] = None,
    ) -> Mapping[str, Any]:
        """
        Specifies request headers.
        Authentication headers will overwrite any overlapping headers returned from this method.
        """
        headers = self._get_request_options(
            stream_state,
            stream_slice,
            next_page_token,
            self.get_request_headers,
            self.get_authenticator().get_auth_header,
            extra_headers,
        )
        if isinstance(headers, str):
            raise ValueError("Request headers cannot be a string")
        return {str(k): str(v) for k, v in headers.items()}

    def _request_params(
        self,
        stream_state: Optional[StreamState],
        stream_slice: Optional[StreamSlice],
        next_page_token: Optional[Mapping[str, Any]],
        extra_params: Optional[Mapping[str, Any]] = None,
    ) -> Mapping[str, Any]:
        """
        Specifies the query parameters that should be set on an outgoing HTTP request given the inputs.

        E.g: you might want to define query parameters for paging if next_page_token is not None.
        """
        options = self._get_request_options(
            stream_state,
            stream_slice,
            next_page_token,
            self.get_request_params,
            self.get_authenticator().get_request_params,
            extra_params,
        )
        if isinstance(options, str):
            raise ValueError("Request params cannot be a string")

        for k, v in options.items():
            if isinstance(v, (dict,)):
                raise ValueError(
                    f"Invalid value for `{k}` parameter. The values of request params cannot be an object."
                )

        return options

    def _request_body_data(
        self,
        stream_state: Optional[StreamState],
        stream_slice: Optional[StreamSlice],
        next_page_token: Optional[Mapping[str, Any]],
        extra_body_data: Optional[Union[Mapping[str, Any], str]] = None,
    ) -> Optional[Union[Mapping[str, Any], str]]:
        """
        Specifies how to populate the body of the request with a non-JSON payload.

        If returns a ready text that it will be sent as is.
        If returns a dict that it will be converted to a urlencoded form.
        E.g. {"key1": "value1", "key2": "value2"} => "key1=value1&key2=value2"

        At the same time only one of the 'request_body_data' and 'request_body_json' functions can be overridden.
        """
        # Warning: use self.state instead of the stream_state passed as argument!
        return self._get_request_options(
            stream_state,
            stream_slice,
            next_page_token,
            self.get_request_body_data,
            self.get_authenticator().get_request_body_data,
            extra_body_data,
        )

    def _request_body_json(
        self,
        stream_state: Optional[StreamState],
        stream_slice: Optional[StreamSlice],
        next_page_token: Optional[Mapping[str, Any]],
        extra_body_json: Optional[Mapping[str, Any]] = None,
    ) -> Optional[Mapping[str, Any]]:
        """
        Specifies how to populate the body of the request with a JSON payload.

        At the same time only one of the 'request_body_data' and 'request_body_json' functions can be overridden.
        """
        # Warning: use self.state instead of the stream_state passed as argument!
        options = self._get_request_options(
            stream_state,
            stream_slice,
            next_page_token,
            self.get_request_body_json,
            self.get_authenticator().get_request_body_json,
            extra_body_json,
        )
        if isinstance(options, str):
            raise ValueError("Request body json cannot be a string")
        return options

    @classmethod
    def _join_url(cls, url_base: str, path: Optional[str] = None) -> str:
        """
        Joins a base URL with a given path and returns the resulting URL with any trailing slash removed.

        This method ensures that there are no duplicate slashes when concatenating the base URL and the path,
        which is useful when the full URL is provided from an interpolation context.

        Args:
            url_base (str): The base URL to which the path will be appended.
            path (Optional[str]): The path to join with the base URL.

        Returns:
            str: The resulting joined URL.

        Note:
            Related issue: https://github.com/airbytehq/airbyte-internal-issues/issues/11869
            - If the path is an empty string or None, the method returns the base URL with any trailing slash removed.

        Example:
            1) _join_url("https://example.com/api/", "endpoint") >> 'https://example.com/api/endpoint'
            2) _join_url("https://example.com/api", "/endpoint") >> 'https://example.com/api/endpoint'
            3) _join_url("https://example.com/api/", "") >> 'https://example.com/api/'
            4) _join_url("https://example.com/api", None) >> 'https://example.com/api'
        """

        # return a full-url if provided directly from interpolation context
        if path == EmptyString or path is None:
            return url_base
        else:
            # since we didn't provide a full-url, the url_base might not have a trailing slash
            # so we join the url_base and path correctly
            if not url_base.endswith("/"):
                url_base += "/"

        return urljoin(url_base, path)

    def send_request(
        self,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
        path: Optional[str] = None,
        request_headers: Optional[Mapping[str, Any]] = None,
        request_params: Optional[Mapping[str, Any]] = None,
        request_body_data: Optional[Union[Mapping[str, Any], str]] = None,
        request_body_json: Optional[Mapping[str, Any]] = None,
        log_formatter: Optional[Callable[[requests.Response], Any]] = None,
    ) -> Optional[requests.Response]:
        request, response = self._http_client.send_request(
            http_method=self.get_method().value,
            url=self._get_url(
                path=path,
                stream_state=stream_state,
                stream_slice=stream_slice,
                next_page_token=next_page_token,
            ),
            request_kwargs={"stream": self.stream_response},
            headers=self._request_headers(
                stream_state, stream_slice, next_page_token, request_headers
            ),
            params=self._request_params(
                stream_state, stream_slice, next_page_token, request_params
            ),
            json=self._request_body_json(
                stream_state, stream_slice, next_page_token, request_body_json
            ),
            data=self._request_body_data(
                stream_state, stream_slice, next_page_token, request_body_data
            ),
            dedupe_query_params=True,
            log_formatter=log_formatter,
            exit_on_rate_limit=self._exit_on_rate_limit,
        )

        return response



import hashlib
from dataclasses import asdict
from typing import Any, Dict, List, Mapping, Optional

import jsonschema
from airbyte_protocol_dataclasses.models import AirbyteStateMessage, ConfiguredAirbyteCatalog
from fastapi import APIRouter, Depends, HTTPException

from airbyte_cdk.models import AirbyteStateMessageSerializer
from airbyte_cdk.sources.declarative.concurrent_declarative_source import (
    ConcurrentDeclarativeSource,
)
from airbyte_cdk.sources.declarative.parsers.custom_code_compiler import (
    INJECTED_COMPONENTS_PY,
    INJECTED_COMPONENTS_PY_CHECKSUMS,
)
from airbyte_cdk.utils.airbyte_secrets_utils import filter_secrets

from ..api_models import (
    CheckRequest,
    CheckResponse,
    DiscoverRequest,
    DiscoverResponse,
    ErrorResponse,
    FullResolveRequest,
    Manifest,
    ManifestResponse,
    ResolveRequest,
    StreamReadResponse,
    StreamTestReadRequest,
)
from ..command_processor.processor import ManifestCommandProcessor
from ..command_processor.utils import build_catalog, build_source
from ..helpers.auth import verify_jwt_token
from ..helpers.tracing import apply_trace_tags_from_context


def safe_build_source(
    manifest_dict: Dict[str, Any],
    config_dict: Mapping[str, Any],
    catalog: Optional[ConfiguredAirbyteCatalog] = None,
    state: Optional[List[AirbyteStateMessage]] = None,
    page_limit: Optional[int] = None,
    slice_limit: Optional[int] = None,
    record_limit: Optional[int] = None,
) -> ConcurrentDeclarativeSource:
    """Wrapper around build_source that converts ValidationError to HTTPException."""
    try:
        return build_source(
            manifest_dict,
            catalog,
            config_dict,
            state,
            record_limit,
            page_limit,
            slice_limit,
        )
    except jsonschema.exceptions.ValidationError as e:
        raise HTTPException(status_code=400, detail=f"Invalid manifest: {e.message}")


router = APIRouter(
    prefix="/manifest",
    tags=["manifest"],
    dependencies=[Depends(verify_jwt_token)],
)


@router.post(
    "/test_read",
    operation_id="testRead",
    responses={
        400: {"description": "Bad Request - Error processing request", "model": ErrorResponse}
    },
)
def test_read(request: StreamTestReadRequest) -> StreamReadResponse:
    """
    Test reading from a specific stream in the manifest.
    """
    # Apply trace tags from context if provided
    if request.context:
        apply_trace_tags_from_context(
            workspace_id=request.context.workspace_id,
            project_id=request.context.project_id,
        )

    config_dict = request.config.model_dump()

    catalog = build_catalog(request.stream_name)
    converted_state = [AirbyteStateMessageSerializer.load(state) for state in request.state]

    if request.custom_components_code:
        config_dict[INJECTED_COMPONENTS_PY] = request.custom_components_code
        config_dict[INJECTED_COMPONENTS_PY_CHECKSUMS] = {
            "md5": hashlib.md5(request.custom_components_code.encode()).hexdigest()
        }

    # We enforce a concurrency level of 1 so that the stream is processed on a single thread
    # to retain ordering for the grouping of the builder message responses.
    manifest = request.manifest.model_dump()
    if "concurrency_level" in manifest:
        manifest["concurrency_level"]["default_concurrency"] = 1
    else:
        manifest["concurrency_level"] = {
            "type": "ConcurrencyLevel",
            "default_concurrency": 1,
        }

    source = safe_build_source(
        manifest,
        config_dict,
        catalog,
        converted_state,
        request.page_limit,
        request.slice_limit,
        request.record_limit,
    )

    runner = ManifestCommandProcessor(source)
    try:
        cdk_result = runner.test_read(
            config_dict,
            catalog,
            converted_state,
            request.record_limit,
            request.page_limit,
            request.slice_limit,
        )
        return StreamReadResponse.model_validate(asdict(cdk_result))
    except Exception as exc:
        # Filter secrets from error message before returning to client
        sanitized_message = filter_secrets(f"Error reading stream: {str(exc)}")
        raise HTTPException(status_code=400, detail=sanitized_message)


@router.post(
    "/check",
    operation_id="check",
    responses={
        400: {"description": "Bad Request - Error processing request", "model": ErrorResponse}
    },
)
def check(request: CheckRequest) -> CheckResponse:
    """Check configuration against a manifest"""
    # Apply trace tags from context if provided
    if request.context:
        apply_trace_tags_from_context(
            workspace_id=request.context.workspace_id,
            project_id=request.context.project_id,
        )

    try:
        source = safe_build_source(request.manifest.model_dump(), request.config.model_dump())
        runner = ManifestCommandProcessor(source)
        success, message = runner.check_connection(request.config.model_dump())
        return CheckResponse(success=success, message=message)
    except Exception as exc:
        # Filter secrets from error message before returning to client
        sanitized_message = filter_secrets(f"Error checking connection: {str(exc)}")
        raise HTTPException(status_code=400, detail=sanitized_message)


@router.post(
    "/discover",
    operation_id="discover",
    responses={
        400: {"description": "Bad Request - Error processing request", "model": ErrorResponse}
    },
)
def discover(request: DiscoverRequest) -> DiscoverResponse:
    """Discover streams from a manifest"""
    # Apply trace tags from context if provided
    if request.context:
        apply_trace_tags_from_context(
            workspace_id=request.context.workspace_id,
            project_id=request.context.project_id,
        )

    try:
        source = safe_build_source(request.manifest.model_dump(), request.config.model_dump())
        runner = ManifestCommandProcessor(source)
        catalog = runner.discover(request.config.model_dump())
        if catalog is None:
            raise HTTPException(
                status_code=422, detail="Connector did not return a discovered catalog"
            )
        return DiscoverResponse(catalog=catalog)
    except HTTPException:
        # Re-raise HTTPExceptions as-is (like the catalog None check above)
        raise
    except Exception as exc:
        # Filter secrets from error message before returning to client
        sanitized_message = filter_secrets(f"Error discovering streams: {str(exc)}")
        raise HTTPException(status_code=400, detail=sanitized_message)


@router.post(
    "/resolve",
    operation_id="resolve",
    responses={
        400: {"description": "Bad Request - Error processing request", "model": ErrorResponse}
    },
)
def resolve(request: ResolveRequest) -> ManifestResponse:
    """Resolve a manifest to its final configuration."""
    # Apply trace tags from context if provided
    if request.context:
        apply_trace_tags_from_context(
            workspace_id=request.context.workspace_id,
            project_id=request.context.project_id,
        )

    try:
        source = safe_build_source(request.manifest.model_dump(), {})
        return ManifestResponse(manifest=Manifest(**source.resolved_manifest))
    except Exception as exc:
        # Filter secrets from error message before returning to client
        sanitized_message = filter_secrets(f"Error resolving manifest: {str(exc)}")
        raise HTTPException(status_code=400, detail=sanitized_message)


@router.post(
    "/full_resolve",
    operation_id="fullResolve",
    responses={
        400: {"description": "Bad Request - Error processing request", "model": ErrorResponse}
    },
)
def full_resolve(request: FullResolveRequest) -> ManifestResponse:
    """
    Fully resolve a manifest, including dynamic streams.

    This is a similar operation to resolve, but has an extra step which generates streams from dynamic stream templates if the manifest contains any. This is used when a user clicks the generate streams button on a stream template in the Builder UI
    """
    # Apply trace tags from context if provided
    if request.context:
        apply_trace_tags_from_context(
            workspace_id=request.context.workspace_id,
            project_id=request.context.project_id,
        )

    try:
        source = safe_build_source(request.manifest.model_dump(), request.config.model_dump())
        manifest = {**source.resolved_manifest}
        streams = manifest.get("streams", [])
        for stream in streams:
            stream["dynamic_stream_name"] = None

        mapped_streams: Dict[str, List[Dict[str, Any]]] = {}
        for stream in source.dynamic_streams:
            generated_streams = mapped_streams.setdefault(stream["dynamic_stream_name"], [])

            if len(generated_streams) < request.stream_limit:
                generated_streams += [stream]

        for generated_streams_list in mapped_streams.values():
            streams.extend(generated_streams_list)

        manifest["streams"] = streams
        return ManifestResponse(manifest=Manifest(**manifest))
    except Exception as exc:
        # Filter secrets from error message before returning to client
        sanitized_message = filter_secrets(f"Error full resolving manifest: {str(exc)}")
        raise HTTPException(status_code=400, detail=sanitized_message)



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from airbyte_cdk.sources.declarative.retrievers.async_retriever import AsyncRetriever
from airbyte_cdk.sources.declarative.retrievers.retriever import Retriever
from airbyte_cdk.sources.declarative.retrievers.simple_retriever import (
    LazySimpleRetriever,
    SimpleRetriever,
)

__all__ = [
    "Retriever",
    "SimpleRetriever",
    "AsyncRetriever",
    "LazySimpleRetriever",
]



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.
"""Run acceptance tests in PyTest.

These tests leverage the same `acceptance-test-config.yml` configuration files as the
acceptance tests in CAT, but they run in PyTest instead of CAT. This allows us to run
the acceptance tests in the same local environment as we are developing in, speeding
up iteration cycles.
"""

from __future__ import annotations

from enum import Enum, auto


class ExpectedOutcome(Enum):
    """Enum to represent the expected outcome of a test scenario.

    Class supports comparisons to a boolean or None.
    """

    EXPECT_EXCEPTION = auto()
    EXPECT_SUCCESS = auto()
    ALLOW_ANY = auto()

    @classmethod
    def from_status_str(cls, status: str | None) -> ExpectedOutcome:
        """Convert a status string to an ExpectedOutcome."""
        if status is None:
            return ExpectedOutcome.ALLOW_ANY

        try:
            return {
                "succeed": ExpectedOutcome.EXPECT_SUCCESS,
                "failed": ExpectedOutcome.EXPECT_EXCEPTION,
                "exception": ExpectedOutcome.EXPECT_EXCEPTION,  # same as 'failed'
            }[status]
        except KeyError as ex:
            raise ValueError(
                f"Invalid status '{status}'. Expected 'succeed', 'failed', or 'exception'.",
            ) from ex

    @classmethod
    def from_expecting_exception_bool(cls, expecting_exception: bool | None) -> ExpectedOutcome:
        """Convert a boolean indicating whether an exception is expected to an ExpectedOutcome."""
        if expecting_exception is None:
            # Align with legacy behavior where default would be 'False' (no exception expected)
            return ExpectedOutcome.EXPECT_SUCCESS

        return (
            ExpectedOutcome.EXPECT_EXCEPTION
            if expecting_exception
            else ExpectedOutcome.EXPECT_SUCCESS
        )

    def expect_exception(self) -> bool:
        """Return whether the expectation is that an exception should be raised."""
        return self == ExpectedOutcome.EXPECT_EXCEPTION

    def expect_success(self) -> bool:
        """Return whether the expectation is that the test should succeed without exceptions."""
        return self == ExpectedOutcome.EXPECT_SUCCESS



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import json
from copy import deepcopy
from enum import Enum
from functools import total_ordering
from typing import Any, Dict, List, Literal, Mapping, Optional, Tuple, Type, Union

from airbyte_cdk.sources.file_based.exceptions import (
    ConfigValidationError,
    FileBasedSourceError,
    SchemaInferenceError,
)

JsonSchemaSupportedType = Union[List[str], Literal["string"], str]
SchemaType = Mapping[str, Mapping[str, JsonSchemaSupportedType]]

schemaless_schema = {"type": "object", "properties": {"data": {"type": "object"}}}

file_transfer_schema = {
    "type": "object",
    "properties": {
        "folder": {"type": "string"},
        "file_name": {"type": "string"},
        "source_uri": {"type": "string"},
        "bytes": {"type": "integer"},
        "id": {"type": ["null", "string"]},
        "created_at": {"type": ["null", "string"]},
        "updated_at": {"type": ["null", "string"]},
        "mime_type": {"type": ["null", "string"]},
    },
}


@total_ordering
class ComparableType(Enum):
    NULL = 0
    BOOLEAN = 1
    INTEGER = 2
    NUMBER = 3
    STRING = 4
    OBJECT = 5

    def __lt__(self, other: Any) -> bool:
        if self.__class__ is other.__class__:
            return self.value < other.value  # type: ignore
        else:
            return NotImplemented


TYPE_PYTHON_MAPPING: Mapping[str, Tuple[str, Optional[Type[Any]]]] = {
    "null": ("null", None),
    "array": ("array", list),
    "boolean": ("boolean", bool),
    "float": ("number", float),
    "integer": ("integer", int),
    "number": ("number", float),
    "object": ("object", dict),
    "string": ("string", str),
}
PYTHON_TYPE_MAPPING = {t: k for k, (_, t) in TYPE_PYTHON_MAPPING.items()}


def get_comparable_type(value: Any) -> Optional[ComparableType]:
    if value == "null":
        return ComparableType.NULL
    if value == "boolean":
        return ComparableType.BOOLEAN
    if value == "integer":
        return ComparableType.INTEGER
    if value == "number":
        return ComparableType.NUMBER
    if value == "string":
        return ComparableType.STRING
    if value == "object":
        return ComparableType.OBJECT
    else:
        return None


def get_inferred_type(value: Any) -> Optional[ComparableType]:
    if value is None:
        return ComparableType.NULL
    if isinstance(value, bool):
        return ComparableType.BOOLEAN
    if isinstance(value, int):
        return ComparableType.INTEGER
    if isinstance(value, float):
        return ComparableType.NUMBER
    if isinstance(value, str):
        return ComparableType.STRING
    if isinstance(value, dict):
        return ComparableType.OBJECT
    else:
        return None


def merge_schemas(schema1: SchemaType, schema2: SchemaType) -> SchemaType:
    """
    Returns a new dictionary that contains schema1 and schema2.

    Schemas are merged as follows
    - If a key is in one schema but not the other, add it to the base schema with its existing type.
    - If a key is in both schemas but with different types, use the wider type.
    - If the type is a list in one schema but a different type of element in the other schema, raise an exception.
    - If the type is an object in both schemas but the objects are different raise an exception.
    - If the type is an object in one schema but not in the other schema, raise an exception.

    In other words, we support merging
    - any atomic type with any other atomic type (choose the wider of the two)
    - list with list (union)
    and nothing else.
    """
    for k, t in list(schema1.items()) + list(schema2.items()):
        if not isinstance(t, dict) or "type" not in t or not _is_valid_type(t["type"]):
            raise SchemaInferenceError(FileBasedSourceError.UNRECOGNIZED_TYPE, key=k, type=t)

    merged_schema: Dict[str, Any] = deepcopy(schema1)  # type: ignore  # as of 2023-08-08, deepcopy can copy Mapping
    for k2, t2 in schema2.items():
        t1 = merged_schema.get(k2)
        if t1 is None:
            merged_schema[k2] = t2
        elif t1 == t2:
            continue
        else:
            merged_schema[k2] = _choose_wider_type(k2, t1, t2)

    return merged_schema


def _is_valid_type(t: JsonSchemaSupportedType) -> bool:
    return t == "array" or get_comparable_type(t) is not None


def _choose_wider_type(key: str, t1: Mapping[str, Any], t2: Mapping[str, Any]) -> Mapping[str, Any]:
    t1_type = t1["type"]
    t2_type = t2["type"]

    if (t1_type == "array" or t2_type == "array") and t1 != t2:
        raise SchemaInferenceError(
            FileBasedSourceError.SCHEMA_INFERENCE_ERROR,
            details="Cannot merge schema for unequal array types.",
            key=key,
            detected_types=f"{t1},{t2}",
        )
    # Schemas can still be merged if a key contains a null value in either t1 or t2, but it is still an object
    elif (
        (t1_type == "object" or t2_type == "object")
        and t1_type != "null"
        and t2_type != "null"
        and t1 != t2
    ):
        raise SchemaInferenceError(
            FileBasedSourceError.SCHEMA_INFERENCE_ERROR,
            details="Cannot merge schema for unequal object types.",
            key=key,
            detected_types=f"{t1},{t2}",
        )
    else:
        comparable_t1 = get_comparable_type(
            TYPE_PYTHON_MAPPING[t1_type][0]
        )  # accessing the type_mapping value
        comparable_t2 = get_comparable_type(
            TYPE_PYTHON_MAPPING[t2_type][0]
        )  # accessing the type_mapping value
        if not comparable_t1 and comparable_t2:
            raise SchemaInferenceError(
                FileBasedSourceError.UNRECOGNIZED_TYPE, key=key, detected_types=f"{t1},{t2}"
            )
        return max(
            [t1, t2],
            key=lambda x: ComparableType(get_comparable_type(TYPE_PYTHON_MAPPING[x["type"]][0])),
        )  # accessing the type_mapping value


def is_equal_or_narrower_type(value: Any, expected_type: str) -> bool:
    if isinstance(value, list):
        # We do not compare lists directly; the individual items are compared.
        # If we hit this condition, it means that the expected type is not
        # compatible with the inferred type.
        return False

    inferred_type = ComparableType(get_inferred_type(value))

    if inferred_type is None:
        return False

    return ComparableType(inferred_type) <= ComparableType(get_comparable_type(expected_type))


def conforms_to_schema(record: Mapping[str, Any], schema: Mapping[str, Any]) -> bool:
    """
    Return true iff the record conforms to the supplied schema.

    The record conforms to the supplied schema iff:
    - All columns in the record are in the schema.
    - For every column in the record, that column's type is equal to or narrower than the same column's
      type in the schema.
    """
    schema_columns = set(schema.get("properties", {}).keys())
    record_columns = set(record.keys())

    if not record_columns.issubset(schema_columns):
        return False

    for column, definition in schema.get("properties", {}).items():
        expected_type = definition.get("type")
        value = record.get(column)

        if value is not None:
            if isinstance(expected_type, list):
                return any(is_equal_or_narrower_type(value, e) for e in expected_type)
            elif expected_type == "object":
                return isinstance(value, dict)
            elif expected_type == "array":
                if not isinstance(value, list):
                    return False
                array_type = definition.get("items", {}).get("type")
                if not all(is_equal_or_narrower_type(v, array_type) for v in value):
                    return False
            elif not is_equal_or_narrower_type(value, expected_type):
                return False

    return True


def _parse_json_input(input_schema: Union[str, Mapping[str, str]]) -> Optional[Mapping[str, str]]:
    try:
        if isinstance(input_schema, str):
            schema: Mapping[str, str] = json.loads(input_schema)
        else:
            schema = input_schema
        if not all(isinstance(s, str) for s in schema.values()):
            raise ConfigValidationError(
                FileBasedSourceError.ERROR_PARSING_USER_PROVIDED_SCHEMA,
                details="Invalid input schema; nested schemas are not supported.",
            )

    except json.decoder.JSONDecodeError:
        return None

    return schema


def type_mapping_to_jsonschema(
    input_schema: Optional[Union[str, Mapping[str, str]]],
) -> Optional[Mapping[str, Any]]:
    """
    Return the user input schema (type mapping), transformed to JSON Schema format.

    Verify that the input schema:
        - is a key:value map
        - all values in the map correspond to a JsonSchema datatype
    """
    if not input_schema:
        return None

    result_schema = {}

    json_mapping = _parse_json_input(input_schema) or {}

    for col_name, type_name in json_mapping.items():
        col_name, type_name = col_name.strip(), type_name.strip()
        if not (col_name and type_name):
            raise ConfigValidationError(
                FileBasedSourceError.ERROR_PARSING_USER_PROVIDED_SCHEMA,
                details=f"Invalid input schema; expected mapping in the format column_name: type, got {input_schema}.",
            )

        _json_schema_type = TYPE_PYTHON_MAPPING.get(type_name.casefold())

        if not _json_schema_type:
            raise ConfigValidationError(
                FileBasedSourceError.ERROR_PARSING_USER_PROVIDED_SCHEMA,
                details=f"Invalid type '{type_name}' for property '{col_name}'.",
            )

        json_schema_type = _json_schema_type[0]
        result_schema[col_name] = {"type": json_schema_type}

    return {"type": "object", "properties": result_schema}



"""
Common API models shared across different endpoints.
"""

from pydantic import BaseModel, ConfigDict


class Manifest(BaseModel):
    """Base manifest model. Allows client generation to replace with proper JsonNode types."""

    model_config = ConfigDict(extra="allow")


class ConnectorConfig(BaseModel):
    """Base connector configuration model. Allows client generation to replace with proper JsonNode types."""

    model_config = ConfigDict(extra="allow")



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#
from abc import abstractmethod
from dataclasses import dataclass
from typing import Any, Iterable, Mapping

import requests


@dataclass
class RecordExtractor:
    """
    Responsible for translating an HTTP response into a list of records by extracting records from the response.
    """

    @abstractmethod
    def extract_records(
        self,
        response: requests.Response,
    ) -> Iterable[Mapping[str, Any]]:
        """
        Selects records from the response
        :param response: The response to extract the records from
        :return: List of Records extracted from the response
        """
        pass



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from abc import ABC, abstractmethod
from typing import Optional


class StreamAvailability:
    @classmethod
    def available(cls) -> "StreamAvailability":
        return cls(True)

    @classmethod
    def unavailable(cls, reason: str) -> "StreamAvailability":
        return cls(False, reason)

    def __init__(self, available: bool, reason: Optional[str] = None) -> None:
        self._available = available
        self._reason = reason

        if not available:
            assert reason, "A reason needs to be provided if the stream is not available"

    @property
    def is_available(self) -> bool:
        """
        :return: True if the stream is available. False if the stream is not
        """
        return self._available

    @property
    def reason(self) -> Optional[str]:
        """
        :return: A message describing why the stream is not available. If the stream is available, this should return None.
        """
        return self._reason



# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
"""Common utilities for manifest server CLI commands."""

import sys

import rich_click as click

# Import server dependencies with graceful fallback
try:
    import ddtrace  # noqa: F401
    import fastapi  # noqa: F401
    import uvicorn  # noqa: F401

    FASTAPI_AVAILABLE = True
except ImportError:
    FASTAPI_AVAILABLE = False


def check_manifest_server_dependencies() -> None:
    """Check if manifest-server dependencies are installed."""
    if not FASTAPI_AVAILABLE:
        click.echo(
            "❌ Manifest runner dependencies not found. Please install with:\n\n"
            "  pip install airbyte-cdk[manifest-server]\n"
            "  # or\n"
            "  poetry install --extras manifest-server\n",
            err=True,
        )
        sys.exit(1)



from typing import Dict

from fastapi import APIRouter

router = APIRouter(
    prefix="/health",
    tags=["health"],
)


@router.get("/")
def health() -> Dict[str, str]:
    return {"status": "ok"}



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Optional

from pydantic.v1 import BaseModel


class RemoteFile(BaseModel):
    """
    A file in a file-based stream.
    """

    uri: str
    last_modified: datetime
    mime_type: Optional[str] = None

    @property
    def file_uri_for_logging(self) -> str:
        """Returns a user-friendly identifier for logging."""
        return self.uri


class UploadableRemoteFile(RemoteFile, ABC):
    """
    A file in a file-based stream that supports uploading(file transferring).
    """

    id: Optional[str] = None
    created_at: Optional[str] = None
    updated_at: Optional[str] = None

    @property
    @abstractmethod
    def size(self) -> int:
        """
        Returns the file size in bytes.
        """
        ...

    @abstractmethod
    def download_to_local_directory(self, local_file_path: str) -> None:
        """
        Download the file from remote source to local storage.
        """
        ...

    @property
    def source_file_relative_path(self) -> str:
        """
        Returns the relative path of the source file.
        """
        return self.uri

    @property
    def source_uri(self) -> str:
        """
        Returns the Source URI for the file being logged.
        """
        return self.uri



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.

from abc import ABC, ABCMeta, abstractmethod
from typing import Any, Iterable

from airbyte_cdk.sources.types import StreamSlice


class StreamSlicerMeta(ABCMeta):
    """
    Metaclass for wrapper scenario that allows it to be used as a type check for StreamSlicer.
    This is necessary because StreamSlicerTestReadDecorator wraps a StreamSlicer and we want to be able to check
    if an instance is a StreamSlicer, even if it is wrapped in a StreamSlicerTestReadDecorator.

    For example in ConcurrentDeclarativeSource, we do things like:
        isinstance(declarative_stream.retriever.stream_slicer,(GlobalSubstreamCursor, PerPartitionWithGlobalCursor))
    """

    def __instancecheck__(cls, instance: Any) -> bool:
        # Check if it's our wrapper with matching wrapped class
        if hasattr(instance, "wrapped_slicer"):
            return isinstance(instance.wrapped_slicer, cls)

        return super().__instancecheck__(instance)


class StreamSlicer(ABC, metaclass=StreamSlicerMeta):
    """
    Slices the stream into chunks that can be fetched independently. Slices enable state checkpointing and data retrieval parallelization.
    """

    @abstractmethod
    def stream_slices(self) -> Iterable[StreamSlice]:
        """
        Defines stream slices

        :return: An iterable of stream slices
        """
        pass



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import json
import logging
from abc import ABC, abstractmethod
from typing import Any, Mapping, Optional

from airbyte_cdk.models import AirbyteLogMessage, AirbyteMessage, Level
from airbyte_cdk.models import Type as MessageType


# Once everything runs on the concurrent CDK and we've cleaned up the legacy flows, we should try to remove
# this class and write messages directly to the message_repository instead of through the logger because for
# cases like the connector builder where ordering of messages is important, using the logger can cause
# messages to be grouped out of order. Alas work for a different day.
class SliceLogger(ABC):
    """
    SliceLogger is an interface that allows us to log slices of data in a uniform way.
    It is responsible for determining whether or not a slice should be logged and for creating the log message.
    """

    SLICE_LOG_PREFIX = "slice:"

    def create_slice_log_message(self, _slice: Optional[Mapping[str, Any]]) -> AirbyteMessage:
        """
        Mapping is an interface that can be implemented in various ways. However, json.dumps will just do a `str(<object>)` if
        the slice is a class implementing Mapping. Therefore, we want to cast this as a dict before passing this to json.dump
        """
        printable_slice = dict(_slice) if _slice else _slice
        return AirbyteMessage(
            type=MessageType.LOG,
            log=AirbyteLogMessage(
                level=Level.INFO,
                message=f"{SliceLogger.SLICE_LOG_PREFIX}{json.dumps(printable_slice, default=str)}",
            ),
        )

    @abstractmethod
    def should_log_slice_message(self, logger: logging.Logger) -> bool:
        """

        :param logger:
        :return:
        """


class DebugSliceLogger(SliceLogger):
    def should_log_slice_message(self, logger: logging.Logger) -> bool:
        """

        :param logger:
        :return:
        """
        return logger.isEnabledFor(logging.DEBUG)


class AlwaysLogSliceLogger(SliceLogger):
    def should_log_slice_message(self, logger: logging.Logger) -> bool:
        return True



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from dataclasses import InitVar, dataclass, field
from typing import Any, Iterable, List, Mapping, MutableMapping, Union

import dpath
import requests

from airbyte_cdk.sources.declarative.decoders import Decoder, JsonDecoder
from airbyte_cdk.sources.declarative.extractors.record_extractor import RecordExtractor
from airbyte_cdk.sources.declarative.interpolation.interpolated_string import InterpolatedString
from airbyte_cdk.sources.types import Config


@dataclass
class DpathExtractor(RecordExtractor):
    """
    Record extractor that searches a decoded response over a path defined as an array of fields.

    If the field path points to an array, that array is returned.
    If the field path points to an object, that object is returned wrapped as an array.
    If the field path points to an empty object, an empty array is returned.
    If the field path points to a non-existing path, an empty array is returned.

    Examples of instantiating this transform:
    ```
      extractor:
        type: DpathExtractor
        field_path:
          - "root"
          - "data"
    ```

    ```
      extractor:
        type: DpathExtractor
        field_path:
          - "root"
          - "{{ parameters['field'] }}"
    ```

    ```
      extractor:
        type: DpathExtractor
        field_path: []
    ```

    Attributes:
        field_path (Union[InterpolatedString, str]): Path to the field that should be extracted
        config (Config): The user-provided configuration as specified by the source's spec
        decoder (Decoder): The decoder responsible to transfom the response in a Mapping
    """

    field_path: List[Union[InterpolatedString, str]]
    config: Config
    parameters: InitVar[Mapping[str, Any]]
    decoder: Decoder = field(default_factory=lambda: JsonDecoder(parameters={}))

    def __post_init__(self, parameters: Mapping[str, Any]) -> None:
        self._field_path = [
            InterpolatedString.create(path, parameters=parameters) for path in self.field_path
        ]
        for path_index in range(len(self.field_path)):
            if isinstance(self.field_path[path_index], str):
                self._field_path[path_index] = InterpolatedString.create(
                    self.field_path[path_index], parameters=parameters
                )

    def extract_records(self, response: requests.Response) -> Iterable[MutableMapping[Any, Any]]:
        for body in self.decoder.decode(response):
            if len(self._field_path) == 0:
                extracted = body
            else:
                path = [path.eval(self.config) for path in self._field_path]
                if "*" in path:
                    extracted = dpath.values(body, path)
                else:
                    extracted = dpath.get(body, path, default=[])  # type: ignore # extracted will be a MutableMapping, given input data structure
            if isinstance(extracted, list):
                yield from extracted
            elif extracted:
                yield extracted
            else:
                yield from []



# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
import logging
import os
from queue import Queue
from typing import Callable, Iterable

from airbyte_cdk.models import AirbyteMessage, Level
from airbyte_cdk.models import Type as MessageType
from airbyte_cdk.sources.message.repository import LogMessage, MessageRepository
from airbyte_cdk.sources.streams.concurrent.partitions.types import QueueItem

logger = logging.getLogger("airbyte")


class ConcurrentMessageRepository(MessageRepository):
    """
    Message repository that immediately loads messages onto the queue processed on the
    main thread. This ensures that messages are processed in the correct order they are
    received. The InMemoryMessageRepository implementation does not have guaranteed
    ordering since whether to process the main thread vs. partitions is non-deterministic
    and there can be a lag between reading the main-thread and consuming messages on the
    MessageRepository.

    This is particularly important for the connector builder which relies on grouping
    of messages to organize request/response, pages, and partitions.
    """

    def __init__(self, queue: Queue[QueueItem], message_repository: MessageRepository):
        self._queue = queue
        self._decorated_message_repository = message_repository

    def emit_message(self, message: AirbyteMessage) -> None:
        self._decorated_message_repository.emit_message(message)
        for message in self._decorated_message_repository.consume_queue():
            self._queue.put(message)

    def log_message(self, level: Level, message_provider: Callable[[], LogMessage]) -> None:
        self._decorated_message_repository.log_message(level, message_provider)
        for message in self._decorated_message_repository.consume_queue():
            self._queue.put(message)

    def consume_queue(self) -> Iterable[AirbyteMessage]:
        """
        This method shouldn't need to be called because as part of emit_message() we are already
        loading messages onto the queue processed on the main thread.
        """
        yield from []



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.

from dataclasses import dataclass
from typing import Any, Optional

from airbyte_cdk.sources.streams.checkpoint import Cursor
from airbyte_cdk.sources.types import Record, StreamSlice, StreamState


@dataclass
class ResumableFullRefreshCursor(Cursor):
    """
    Cursor that allows for the checkpointing of sync progress according to a synthetic cursor based on the pagination state
    of the stream. Resumable full refresh syncs are only intended to retain state in between sync attempts of the same job
    with the platform responsible for removing said state.
    """

    def __init__(self) -> None:
        self._cursor: StreamState = {}

    def get_stream_state(self) -> StreamState:
        return self._cursor

    def set_initial_state(self, stream_state: StreamState) -> None:
        self._cursor = stream_state

    def observe(self, stream_slice: StreamSlice, record: Record) -> None:
        """
        Resumable full refresh manages state using a page number so it does not need to update state by observing incoming records.
        """
        pass

    def close_slice(self, stream_slice: StreamSlice, *args: Any) -> None:
        self._cursor = stream_slice.cursor_slice

    def should_be_synced(self, record: Record) -> bool:
        """
        Unlike date-based cursors which filter out records outside slice boundaries, resumable full refresh records exist within pages
        that don't have filterable bounds. We should always return them.
        """
        return True

    def select_state(self, stream_slice: Optional[StreamSlice] = None) -> Optional[StreamState]:
        # A top-level RFR cursor only manages the state of a single partition
        return self._cursor



# Copyright (c) 2025 Airbyte, Inc., all rights reserved.

from airbyte_cdk.sources.declarative.requesters.query_properties.property_selector.json_schema_property_selector import (
    JsonSchemaPropertySelector,
)
from airbyte_cdk.sources.declarative.requesters.query_properties.property_selector.property_selector import (
    PropertySelector,
)

__all__ = ["JsonSchemaPropertySelector", "PropertySelector"]



import copy
from typing import Any, Dict, List, Mapping, Optional

from airbyte_protocol_dataclasses.models import AirbyteStateMessage

from airbyte_cdk.models import (
    AirbyteStream,
    ConfiguredAirbyteCatalog,
    ConfiguredAirbyteStream,
    DestinationSyncMode,
    SyncMode,
)
from airbyte_cdk.sources.declarative.concurrent_declarative_source import (
    ConcurrentDeclarativeSource,
    TestLimits,
)

SHOULD_NORMALIZE_KEY = "__should_normalize"
SHOULD_MIGRATE_KEY = "__should_migrate"


def build_catalog(stream_name: str) -> ConfiguredAirbyteCatalog:
    return ConfiguredAirbyteCatalog(
        streams=[
            ConfiguredAirbyteStream(
                stream=AirbyteStream(
                    name=stream_name,
                    json_schema={},
                    supported_sync_modes=[SyncMode.full_refresh, SyncMode.incremental],
                ),
                sync_mode=SyncMode.incremental,
                destination_sync_mode=DestinationSyncMode.overwrite,
            )
        ]
    )


def should_migrate_manifest(manifest: Mapping[str, Any]) -> bool:
    """
    Determines whether the manifest should be migrated,
    based on the presence of the "__should_migrate" key.

    This flag is set by the UI.
    """
    return manifest.get(SHOULD_MIGRATE_KEY, False)


def should_normalize_manifest(manifest: Mapping[str, Any]) -> bool:
    """
    Determines whether the manifest should be normalized,
    based on the presence of the "__should_normalize" key.

    This flag is set by the UI.
    """
    return manifest.get(SHOULD_NORMALIZE_KEY, False)


def build_source(
    manifest: Dict[str, Any],
    catalog: Optional[ConfiguredAirbyteCatalog],
    config: Mapping[str, Any],
    state: Optional[List[AirbyteStateMessage]],
    record_limit: Optional[int] = None,
    page_limit: Optional[int] = None,
    slice_limit: Optional[int] = None,
) -> ConcurrentDeclarativeSource:
    definition = copy.deepcopy(manifest)

    should_normalize = should_normalize_manifest(manifest)
    if should_normalize:
        del definition[SHOULD_NORMALIZE_KEY]

    should_migrate = should_migrate_manifest(manifest)
    if should_migrate:
        del definition[SHOULD_MIGRATE_KEY]

    return ConcurrentDeclarativeSource(
        catalog=catalog,
        state=state,
        source_config=definition,
        config=config,
        normalize_manifest=should_normalize,
        migrate_manifest=should_migrate,
        emit_connector_builder_messages=True,
        limits=TestLimits(
            max_pages_per_slice=page_limit or TestLimits.DEFAULT_MAX_PAGES_PER_SLICE,
            max_slices=slice_limit or TestLimits.DEFAULT_MAX_SLICES,
            max_records=record_limit or TestLimits.DEFAULT_MAX_RECORDS,
        ),
    )



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#


from typing import Any, Dict, Iterator, List, Mapping, Optional

from airbyte_cdk.connector_builder.models import (
    AuxiliaryRequest,
    HttpRequest,
    HttpResponse,
    StreamReadPages,
)
from airbyte_cdk.models import (
    AirbyteMessage,
)
from airbyte_cdk.utils.datetime_format_inferrer import DatetimeFormatInferrer
from airbyte_cdk.utils.schema_inferrer import (
    SchemaInferrer,
)

from .helpers import (
    airbyte_message_to_json,
    handle_current_page,
    handle_current_slice,
    handle_log_message,
    handle_record_message,
    is_async_auxiliary_request,
    is_config_update_message,
    is_log_message,
    is_page_http_request_for_different_stream,
    is_record_message,
    is_state_message,
    is_trace_with_error,
    parse_slice_description,
    should_close_page,
    should_close_page_for_slice,
    should_process_slice_descriptor,
)
from .types import MESSAGE_GROUPS


def get_message_groups(
    messages: Iterator[AirbyteMessage],
    schema_inferrer: SchemaInferrer,
    datetime_format_inferrer: DatetimeFormatInferrer,
    limit: int,
    stream_name: str,
) -> MESSAGE_GROUPS:
    """
    Processes an iterator of AirbyteMessage objects to group and yield messages based on their type and sequence.

    This function iterates over the provided messages until the number of record messages processed reaches the specified limit.
    It accumulates messages into pages and slices, handling various types of messages such as log, trace (with errors), record,
    configuration update, and state messages. The function makes use of helper routines to:
        - Convert messages to JSON.
        - Determine when to close a page or a slice.
        - Parse slice descriptors.
        - Handle log messages and auxiliary requests.
        - Process record messages while inferring schema and datetime formats.

    Depending on the incoming message type, it may yield:
        - StreamReadSlices objects when a slice is completed.
        - Auxiliary HTTP requests/responses generated from log messages.
        - Error trace messages if encountered.
        - Configuration update messages.

    Parameters:
        messages (Iterator[AirbyteMessage]): An iterator yielding AirbyteMessage instances.
        schema_inferrer (SchemaInferrer): An instance used to infer and update schema based on record messages.
        datetime_format_inferrer (DatetimeFormatInferrer): An instance used to infer datetime formats from record messages.
        limit (int): The maximum number of record messages to process before stopping.

    Yields:
        Depending on the type of message processed, one or more of the following:
            - StreamReadSlices: A grouping of pages within a slice along with the slice descriptor and state.
            - HttpRequest/HttpResponse: Auxiliary request/response information derived from log messages.
            - TraceMessage: Error details when a trace message with errors is encountered.
            - ControlMessage: Configuration update details.

    Notes:
        The function depends on several helper functions (e.g., airbyte_message_to_json, should_close_page,
        handle_current_page, parse_slice_description, handle_log_message, and others) to encapsulate specific behavior.
        It maintains internal state for grouping pages and slices, ensuring that related messages are correctly aggregated
        and yielded as complete units.
    """

    records_count = 0
    at_least_one_page_in_group = False
    current_page_records: List[Mapping[str, Any]] = []
    current_slice_descriptor: Optional[Dict[str, Any]] = None
    current_slice_pages: List[StreamReadPages] = []
    current_page_request: Optional[HttpRequest] = None
    current_page_response: Optional[HttpResponse] = None
    latest_state_message: Optional[Dict[str, Any]] = None
    slice_auxiliary_requests: List[AuxiliaryRequest] = []

    while message := next(messages, None):
        json_message = airbyte_message_to_json(message)

        if is_page_http_request_for_different_stream(json_message, stream_name):
            continue

        if should_close_page(at_least_one_page_in_group, message, json_message):
            current_page_request, current_page_response = handle_current_page(
                current_page_request,
                current_page_response,
                current_slice_pages,
                current_page_records,
            )

        if should_close_page_for_slice(at_least_one_page_in_group, message):
            yield handle_current_slice(
                current_slice_pages,
                current_slice_descriptor,
                latest_state_message,
                slice_auxiliary_requests,
            )
            current_slice_descriptor = parse_slice_description(message.log.message)  # type: ignore
            current_slice_pages = []
            at_least_one_page_in_group = False
        elif should_process_slice_descriptor(message):
            # parsing the first slice
            current_slice_descriptor = parse_slice_description(message.log.message)  # type: ignore
        elif is_log_message(message):
            (
                at_least_one_page_in_group,
                current_page_request,
                current_page_response,
                auxiliary_request,
                log_message,
            ) = handle_log_message(
                message,
                json_message,
                at_least_one_page_in_group,
                current_page_request,
                current_page_response,
            )

            if auxiliary_request:
                if is_async_auxiliary_request(auxiliary_request):
                    slice_auxiliary_requests.append(auxiliary_request)
                else:
                    yield auxiliary_request

            if log_message:
                yield log_message
        elif is_trace_with_error(message):
            if message.trace is not None:
                yield message.trace
        elif is_record_message(message):
            records_count = handle_record_message(
                message,
                schema_inferrer,
                datetime_format_inferrer,
                records_count,
                current_page_records,
            )
        elif is_config_update_message(message):
            if message.control is not None:
                yield message.control
        elif is_state_message(message):
            latest_state_message = message.state  # type: ignore

    else:
        if current_page_request or current_page_response or current_page_records:
            handle_current_page(
                current_page_request,
                current_page_response,
                current_slice_pages,
                current_page_records,
            )
            yield handle_current_slice(
                current_slice_pages,
                current_slice_descriptor,
                latest_state_message,
                slice_auxiliary_requests,
            )



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.

from abc import ABC

from airbyte_cdk.sources.declarative.stream_slicers.stream_slicer import StreamSlicer
from airbyte_cdk.sources.streams.checkpoint.cursor import Cursor


class DeclarativeCursor(Cursor, StreamSlicer, ABC):
    """
    DeclarativeCursors are components that allow for checkpointing syncs. In addition to managing the fetching and updating of
    state, declarative cursors also manage stream slicing and injecting slice values into outbound requests.
    """



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import itertools
import logging
from collections import ChainMap
from collections.abc import Callable
from dataclasses import InitVar, dataclass
from typing import Any, Iterable, List, Mapping, Optional

from airbyte_cdk.sources.declarative.partition_routers.partition_router import PartitionRouter
from airbyte_cdk.sources.declarative.partition_routers.substream_partition_router import (
    SubstreamPartitionRouter,
)
from airbyte_cdk.sources.types import StreamSlice, StreamState


def check_for_substream_in_slicers(
    slicers: Iterable[PartitionRouter], log_warning: Callable[[str], None]
) -> None:
    """
    Recursively checks for the presence of SubstreamPartitionRouter within slicers.
    Logs a warning if a SubstreamPartitionRouter is found within a CartesianProductStreamSlicer.

    Args:
        slicers (Iterable[PartitionRouter]): The list of slicers to check.
        log_warning (Callable): Logging function to record warnings.
    """
    for slicer in slicers:
        if isinstance(slicer, SubstreamPartitionRouter):
            log_warning("Parent state handling is not supported for CartesianProductStreamSlicer.")
            return
        elif isinstance(slicer, CartesianProductStreamSlicer):
            # Recursively check sub-slicers within CartesianProductStreamSlicer
            check_for_substream_in_slicers(slicer.stream_slicers, log_warning)


@dataclass
class CartesianProductStreamSlicer(PartitionRouter):
    """
    Stream slicers that iterates over the cartesian product of input stream slicers
    Given 2 stream slicers with the following slices:
    A: [{"i": 0}, {"i": 1}, {"i": 2}]
    B: [{"s": "hello"}, {"s": "world"}]
    the resulting stream slices are
    [
        {"i": 0, "s": "hello"},
        {"i": 0, "s": "world"},
        {"i": 1, "s": "hello"},
        {"i": 1, "s": "world"},
        {"i": 2, "s": "hello"},
        {"i": 2, "s": "world"},
    ]

    Attributes:
        stream_slicers (List[PartitionRouter]): Underlying stream slicers. The RequestOptions (e.g: Request headers, parameters, etc..) returned by this slicer are the combination of the RequestOptions of its input slicers. If there are conflicts e.g: two slicers define the same header or request param, the conflict is resolved by taking the value from the first slicer, where ordering is determined by the order in which slicers were input to this composite slicer.
    """

    stream_slicers: List[PartitionRouter]
    parameters: InitVar[Mapping[str, Any]]

    def __post_init__(self, parameters: Mapping[str, Any]) -> None:
        check_for_substream_in_slicers(self.stream_slicers, self.logger.warning)

    def get_request_params(
        self,
        *,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> Mapping[str, Any]:
        return dict(
            ChainMap(
                *[  # type: ignore # ChainMap expects a MutableMapping[Never, Never] for reasons
                    s.get_request_params(
                        stream_state=stream_state,
                        stream_slice=stream_slice,
                        next_page_token=next_page_token,
                    )
                    for s in self.stream_slicers
                ]
            )
        )

    def get_request_headers(
        self,
        *,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> Mapping[str, Any]:
        return dict(
            ChainMap(
                *[  # type: ignore # ChainMap expects a MutableMapping[Never, Never] for reasons
                    s.get_request_headers(
                        stream_state=stream_state,
                        stream_slice=stream_slice,
                        next_page_token=next_page_token,
                    )
                    for s in self.stream_slicers
                ]
            )
        )

    def get_request_body_data(
        self,
        *,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> Mapping[str, Any]:
        return dict(
            ChainMap(
                *[  # type: ignore # ChainMap expects a MutableMapping[Never, Never] for reasons
                    s.get_request_body_data(
                        stream_state=stream_state,
                        stream_slice=stream_slice,
                        next_page_token=next_page_token,
                    )
                    for s in self.stream_slicers
                ]
            )
        )

    def get_request_body_json(
        self,
        *,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> Mapping[str, Any]:
        return dict(
            ChainMap(
                *[  # type: ignore # ChainMap expects a MutableMapping[Never, Never] for reasons
                    s.get_request_body_json(
                        stream_state=stream_state,
                        stream_slice=stream_slice,
                        next_page_token=next_page_token,
                    )
                    for s in self.stream_slicers
                ]
            )
        )

    def stream_slices(self) -> Iterable[StreamSlice]:
        sub_slices = (s.stream_slices() for s in self.stream_slicers)
        product = itertools.product(*sub_slices)
        for stream_slice_tuple in product:
            partition = dict(ChainMap(*[s.partition for s in stream_slice_tuple]))  # type: ignore # ChainMap expects a MutableMapping[Never, Never] for reasons
            cursor_slices = [s.cursor_slice for s in stream_slice_tuple if s.cursor_slice]
            extra_fields = dict(ChainMap(*[s.extra_fields for s in stream_slice_tuple]))  # type: ignore # ChainMap expects a MutableMapping[Never, Never] for reasons
            if len(cursor_slices) > 1:
                raise ValueError(
                    f"There should only be a single cursor slice. Found {cursor_slices}"
                )
            if cursor_slices:
                cursor_slice = cursor_slices[0]
            else:
                cursor_slice = {}
            yield StreamSlice(
                partition=partition, cursor_slice=cursor_slice, extra_fields=extra_fields
            )

    def get_stream_state(self) -> Optional[Mapping[str, StreamState]]:
        """
        Parent stream states are not supported for cartesian product stream slicer
        """
        pass

    @property
    def logger(self) -> logging.Logger:
        return logging.getLogger("airbyte.CartesianProductStreamSlicer")
